# script de définition des associations des programmes par défaut
# lancé au logon pour l'utilisateur en cours

$Path = "C:\netinst\"
If(!(Test-Path -PathType Container $Path)) {
    New-Item -ItemType Directory -Path $Path
}

Import-Module "$env:ProgramFiles\Sambaedu\SFTA.ps1"

Get-Date -Format "yyyy-MM-dd HH:mm:ss" | Out-File -FilePath "C:\netinst\associations.log" -Force

# récupération de la liste
$list = @{}
Try {
    $list.file = Get-FTA
    #Write-Output $list
} Catch {
    Write-Host "Failed to list default files associations..." -ForegroundColor Red | Out-File -FilePath "C:\netinst\associations.log" -Append
    Exit 1
}

Try {
    $list.protocol = Get-PTA
    #Write-Output $list
} Catch {
    Write-Host "Failed to list default protocols associations..." -ForegroundColor Red | Out-File -FilePath "C:\netinst\associations.log" -Append
    Exit 1
}


$Uri = "http://${env:SE4FS}/gpo/associations_out.php"
$Form = @{
    id   = $env:ID
    list = $list | ConvertTo-JSON
}
Write-Host "$uri $env:id" | Out-File -FilePath "C:\netinst\associations.log" -Append
Try {
    $associations = Invoke-RestMethod -Uri $Uri -Method Post -Body $Form
} Catch {
    Write-Host "Failed to upload default associations... " -ForegroundColor Red | Out-File -FilePath "C:\netinst\associations.log" -Append
    Exit 1
}
$result = $associations.result
If ($result) {
    ForEach ( $a in $result.PSobject.Properties ) {
        $index = $($a.Name)
        $val = $($a.Value)
        Write-Host "$index : $($val.type) $($val.ProgId)" | Out-File -FilePath "C:\netinst\associations.log" -Append
        If ($val.type -eq "file"){
            Set-FTA $($val.ProgId) $index
        } Else {
            Set-PTA $($val.ProgId) $index
        }
    }
}
